/**************************************************************************************************
 *                                                                                                *
 *            Przyklad wykorzystania klasy hash_map, autor: Maciej Zbierski, H6ISI                *
 *                                                                                                *
 **************************************************************************************************
 * Informacje od autora: Klasa hash_map jest aktualnie zastpowana przez klas unordered_map,     *
 * dostepn w standardzie tr1. Praktycznie jedyne rnice midzy tymi klasami dotycz sposobu     *
 * zmiany liczby kubekw, w ktrych klasa przechowuje powierzone jej informacje. Klasa hash_map  *
 * nie jest obsugiwana przez kompilatory g++ od wersji 4.0, w MS Visual C++, od wersji 7.1       *
 * umieszczona jest w namespace stdext, czyli oznaczona jako przestarzaa. Poniszy przykad      *
 * ilustruje wykorzystanie kontenera hash_map i powinien by uruchamiany pod kompilatorem         *
 * MS Visual C++. Do przykadu doczono odpowiednie deklaracje makrogeneratora umoliwiajce     *
 * skompilowanie tego programu pod innymi kompilatorami (przy uyciu kontenera unordered_map ze   *
 * standardu tr1).                                                                                *
 **************************************************************************************************/

#include <iostream>
#include <string>

#ifdef _MSC_VER																	// Visual C++
	#include <hash_map>	
	using namespace stdext;
#else																			// inne kompilatory
	#include <tr1/unordered_map>
	#define hash_map tr1::unordered_map
#endif

using namespace std;

// funkcja pomocnicza wyswietlajaca rozmiar i zawartosc mapy haszujcej
void wyswietlMiesiace(hash_map<string, int>& miesiace)
{
	// wywietl rozmiar kontenera
	cout << "W roku mamy " << miesiace.size() << " miesiecy. Oto one:" << endl;

	// wywietl wszystkie elementy
	for (hash_map<string, int>::const_iterator iter = miesiace.begin(); iter != miesiace.end(); ++iter)
		cout << iter->first << " ma " << iter->second << " dni." << endl;

	cout << endl;
}

/* funkcja pomocnicza wyswietlajca informacje na temat rozmieszczenia elementw w mapie haszujcej, czyli
 *  - liczb kubekw, w ktrych znajduj si elementy 
 *  - load factor, czyli stosunek liczby elementw w hashmapie do liczby kubekw (innymi sowy: ile maksymalnie elementw bdzie w kadym kubeku)
 *  - rozmiar mapy haszujcej - aktualny i maksymalny */
void wyswietlInformacjeTechniczne(hash_map<string, int>& miesiace)
{
	cout << "Mapa haszujaca ma nastepujaca liczbe kubelkow: " << miesiace.bucket_count() << endl;
	cout << "Aktualny load factor to: " << miesiace.load_factor() << ", a maksymalny to " << miesiace.max_load_factor() << endl;
	cout << "Aktualny rozmiar mapy to " << miesiace.size() << ", a maksymalny to " << miesiace.max_size() << endl << endl;
}

int main()
{
	// utwrz map haszujc
	hash_map<string, int> miesiace;	

	// dodaj przykadowe informacje z wykorzystaniem operatora []
	miesiace["styczen"]     = 31;
	miesiace["luty"]        = 28;
	miesiace["marzec"]      = 31;
	miesiace["kwiecien"]    = 30;
	miesiace["maj"]         = 31;
	miesiace["czerwiec"]    = 30;
	miesiace["lipiec"]      = 31;
	miesiace["sierpien"]    = 31;
	miesiace["wrzesien"]    = 30;
	miesiace["pazdziernik"] = 31;
	miesiace["listopad"]    = 30;
	miesiace["grudzien"]    = 31;

	/* wywietl list miesicy - naley pamita, e mapa haszujca nie przechowuje elementw
	 * w kolejnoci wstawienia, wic miesice nie bd wypisane po kolei */
	wyswietlMiesiace(miesiace);

	// Przyklad dodawania elementw:
	// deklaracja zmiennej, w ktrej umieszczony zostanie rezultat wykonania polecenia insert
	pair<hash_map<string, int>::iterator, bool> wynikDodania;

	// prba dodania elementu o ju istniejcym kluczu, powinna si nie uda, gdy nie jest to multi mapa
	wynikDodania = miesiace.insert(pair<string, int>("luty", 29));
	cout << "Czy dodano kolejny miesiac: " << wynikDodania.second << endl;						

	// prba dodania elementu o kluczu nieistniejcym w mapie
	wynikDodania = miesiace.insert(pair<string, int>("sesja", 14));
	cout << "Czy dodano kolejny miesiac: " << wynikDodania.second << endl << endl;

	// wywietlenie wszystkich miesicy wcznie z nowododanym
	wyswietlMiesiace(miesiace);

	/* Przykad zmiany ju istniejcych elementw. Mona to zrobi na dwa sposoby:
	   1) Poprzez operator [] - elementy o kluczu nieistniejcym w mapie zostan utworzone
	   2) Poprzez funkcj find zwracajc end(), gdy element nie znajduje si w mapie
	      lub referencj do obiektu w przeciwnym przypadku - przykad poniej */
	hash_map<string, int>::iterator znaleziony = miesiace.find("luty");
	if (znaleziony == miesiace.end())
		cout << "Nie znaleziono miesiaca 'luty' " << endl;
	else
		znaleziony->second = 29;

	// wywietlenie miesicy po zmianie iloci dni w miesicu lutym
	wyswietlMiesiace(miesiace);

	// wywietlenie informacji technicznych na temat mapy
	wyswietlInformacjeTechniczne(miesiace);

	/* zmiana maksymalnego load factor na warto 1.0 (maksymalnie 1 element na kady kubeek). 
	 * Jeli dokonuje si zmiany (na niepustej hashmapie) na load factor mniejszy od aktualnego, 
	 * naley wywoa funkcj rehash (o ktrej poniej) z dowolnym parametrem - liczba kubekw
	 * zostanie ustawiona na najmniejsz moliw warto, ktra spenia warunek max_load_factor.
	 * W najlepszym wypadku naley wykonywa t funkcj gdy jeszcze mapa jest pusta 
	 * - w celu uniknicia wywoania kosztownej funkcji rehash() */
	miesiace.max_load_factor(1.0);

	/* funkcja rehash odpowiedzialna jest za zmian liczby kubekw na nie mniejsz ni podana
	 * jako parametr funkcji (moe by wiksza, jeli wymaga tego ustawienie max_load_factor). 
	 * Jest to odperaca do kosztowna, gdy wymaga zmiany wewntrznej struktury przechowywania
	 * danych. */
	miesiace.rehash(32);

	// potwierdzenie wprowadzonych zmian w strukturze mapy
	wyswietlInformacjeTechniczne(miesiace);

	// usunicie pojedyczego wpisu
	miesiace.erase("sesja");

	// potwierdzenie usunicia miesica "sesja"
	wyswietlMiesiace(miesiace);

	// wyczyszczenie caej mapy
	miesiace.clear();

	// potwierdzenie wyczyszczenia
	wyswietlMiesiace(miesiace);
}
